#!/usr/bin/env python

import sys
import struct

class solfile:
	def __init__(self, file):
		self.file = file
		self.totallen = ''
		self.bytesint = [['uw', 'luw', 'd', 'c'], ['!h', '!l', '!d', '!i']]
		self.datatypes = {'number': '\x00','boolean': '\x01','string': '\x02','object': '\x03','null': '\x05','undefined': '\x06','array': '\x08','c_object': '\x10', 'date': '\x0B','xml': '\x0F'}
		self.debug = ''
		self.debugtab = []
		self.filtervalue = ''

	def _add_debug(self, head):
		if self.debug != 'no' and self.debug != '':
			#print repr(self.debug)
			self.debugtab.append('%s: %s' % (head, self.debug))
			self.debug = ''

	def _checkheader(self):
		#unknown 2 bytes
		self._testbytes(2, '\x00\xbf')
		#size (cardinal)
		self.totallen = self._bytes2int(4, 'c')
		self._add_debug('fileheader')
		#TCSO + unknown 6 bytes
		self._testbytes(10, '\x54\x43\x53\x4F\x00\x04\x00\x00\x00\x00')
		#globalname
		self.globalname = self._returnbytes(self._bytes2int(2, 'uw'))
		#unknown block of 4 bytes of 00 sometime: '\x00\x00\x00\x03'
		self._testbytes(4, '\x00\x00\x00\x00', '\x00\x00\x00\x03')
		self._add_debug('header')

	def _createheader(self, globalname):
		self._createbytes('\x54\x43\x53\x4F\x00\x04\x00\x00\x00\x00')
		self._createbytes(str(self._int2bytes(len(globalname), 'uw')))
		self._createbytes(globalname)
		self._createbytes('\x00\x00\x00\x00')
		self._add_debug('header')

	def _returnbytes(self, len):
		if self.totallen != '':
			self.totallen = int(self.totallen)-int(len)
		val = self.solfile.read(len)
		if self.debug != 'no':
			self.debug += val
		return val

	def _createbytes(self, val='', sync='no'):
		if self.debug != 'no':
			self.debug += val
		self.solcontent += val
		return len(val)

	def _bytes2int(self, len, typeval):
		val = self._returnbytes(len)
		ctype = self.bytesint[1][self.bytesint[0].index(typeval)]
		return struct.unpack(ctype, val)[0]

	def _int2bytes(self, val, typeval):
		ctype = self.bytesint[1][self.bytesint[0].index(typeval)]
		return struct.pack(ctype, val)

	def _testbytes(self, len, pattern, notpattern=''):
		issol = self._returnbytes(len)
		if issol != pattern:
			if issol == notpattern:
				print '%s is not a standard solfile' % self.file
			else:
				print repr(self.debug)
				print 'testbytes error: %s - %s for file %s' % (repr(pattern), repr(issol), self.file)
			sys.exit(1)

	def _readcontent(self, init='no'): 
		if init == 'yes':
			name = self.globalname
			datatype = 'object'
		else:
			self._add_debug('content')
			#length of the name
			len = self._bytes2int(2, 'uw')
			if len == 0:
				endbytes = self._returnbytes(1)
				if endbytes == '\x09':
					return 2
			#name
			name = self._returnbytes(len)
			#type
			type = self._returnbytes(1)
			if type == self.datatypes['number']:
				datatype = 'number'
				value = self._bytes2int(8, 'd')
			elif type == self.datatypes['boolean']:
				datatype = 'boolean'
				if self._returnbytes(1) == '\x00':
					value = 'False'
				else:
					value = 'True'
			elif type == self.datatypes['string']:
				datatype = 'string'
				lenght = self._bytes2int(2, 'uw')
				value = self._returnbytes(lenght)
			elif type == self.datatypes['object']:
				datatype = 'object'
			elif type == self.datatypes['null']:
				datatype = 'null'
				value = 'NULL'
		
			elif type == self.datatypes['undefined']:
				datatype = 'undefined'
				value = ''
			elif type == self.datatypes['array']:
				datatype = 'array'
				arraylen = self._bytes2int(4, 'luw')
			elif type == self.datatypes['c_object']:
				datatype = 'c_object'
				lenght = self._bytes2int(2, 'uw')
				classname = self._returnbytes(lenght)
			elif type == self.datatypes['date']:
				datatype = 'date'
				time = self._returnbytes(8)
				timezone = self._returnbytes(2)
				value = [time, timezone]
			elif type == self.datatypes['xml']:
				datatype = 'xml'
				lenght = self._bytes2int(4, 'luw')
				value = self._returnbytes(lenght)
			else:
				print 'unknown type %s : name %s : %s for file %s' % (repr(type), repr(name), repr(self._returnbytes(3)), self.file)
				sys.exit(1)
		if datatype == 'object' or datatype == 'array' or datatype == 'c_object':
			code = ''
			returnagain = {}
			while code != 2:
				code = self._readcontent()
				if code != 2:
					key = code.keys()[0]
					returnagain[key] = code[key]
				if init == 'yes':
					self._testbytes(1, '\x00')
				if self.totallen == 0:
					code = 2
			if datatype == 'c_object':
				returnvalue = {name: [datatype, returnagain, classname]}
			elif datatype == 'array':
				returnvalue = {name: [datatype, returnagain, arraylen]}
			else:
				returnvalue = {name: [datatype, returnagain]}
		else:
			returnvalue = {name: [datatype, value]}
		return returnvalue

	def _createcontent(self, value, init='no'):
		for name in value.keys():
			self._add_debug('content')
			self._createbytes(self._int2bytes(len(name), 'uw'))
			self._createbytes(name)
			typeval = value[name][0]
			self._createbytes(self.datatypes[typeval])
			if typeval == 'number':
				self._createbytes(self._int2bytes(value[name][1], 'd'))
			elif typeval == 'boolean':
				if value[name][1] == 'False':
					self._createbytes('\x00')
				else:
					self._createbytes('\x01')
			elif typeval == 'string':
				self._createbytes(self._int2bytes(len(value[name][1]), 'uw'))
				self._createbytes(value[name][1])
			elif typeval == 'object':
				self._createcontent(value[name][1])
			elif typeval == 'array':
				self._createbytes(self._int2bytes(value[name][2], 'luw'))
				self._createcontent(value[name][1])
			elif typeval == 'c_object':
				classname = value[name][2]
				self._createbytes(self._int2bytes(len(classname), 'uw'))
				self._createbytes(classname)
				self._createcontent(value[name][1])
			elif typeval == 'date':
				self._createbytes(value[name][1][0])
				self._createbytes(value[name][1][1])
			elif typeval == 'xml':
				self._createbytes(self._int2bytes(len(value[name][1]), 'luw'))
				self._createbytes(value[name][1])
			elif not typeval == 'undefined' and not typeval == 'null':
				print 'unsupported: ' + typeval
			if typeval == 'object' or typeval == 'array' or typeval == 'c_object':
				self._add_debug('content')
				self._createbytes('\x00\x00\x09')
			if init == 'yes':
				self._createbytes('\x00')


	def filter(self, filtervalue, filterseparator='_'):
		self.filtervalue = filtervalue.split(filterseparator)

	def get_sol(self):
		#FIXME: verifier s'il existe
		self.solfile=open(self.file, 'rb')
		self._checkheader()
		solcontent = self._readcontent(init='yes')
		self.solfile.close()
		self._add_debug('content')
		listcontent = solcontent
		if self.filtervalue != '':
			for search in self.filtervalue:
				if listcontent.has_key(search):
					listcontent = listcontent[search][1]
				else:
					print '%s does not exist' % self.filtervalue 
					sys.exit(1)

		return listcontent

	def set_sol(self, value, test='no'):
		self.solcontent = ''
		self.solcontenttmp = ''
		globalname = value.keys()[0]
		self._createheader(globalname)
		self._createcontent(value[globalname][1], 'yes')
		self._add_debug('content')
		lenght = len(self.solcontent)
		#start header (with len)
		self.startheader = '\x00\xbf' + self._int2bytes(lenght, 'c')
		self.debug = self.startheader
		self._add_debug('fileheader')
		self.solcontent = self.startheader + self.solcontent
		if test == 'no':
			solfile=open(self.file, 'wb')
			solfile.write(self.solcontent)
			solfile.close()

	def print_debug(self):
		if self.debug == 'no':
			return 'no debug information'
		return self.debugtab

from optparse import OptionParser
parser = OptionParser()
parser.add_option('-t', '--test', action="store_true", dest='testit', default=False,
	help='Test to read the given file and rewrite it to test.sol file for testing')
parser.add_option('-r', '--read', dest='filename',
	help='Read the the given file')
parser.add_option('-c', '--create', action="store_true", dest='createsol', default=False,
    help='Create empty settings file')
parser.add_option('-f', '--filter', dest='filtername',
	help='Add filter (for example : settings_secureCrossDomainCacheSize)')
parser.add_option('-d', '--debug', action="store_true", dest='debugit', default=False,
	help='Print debug information')
parser.add_option('-a', '--add', dest='addsol', help='Add attribute (need --read attribute)'),
parser.add_option('-s', '--suppress', dest='supprsol',  help='Suppress attribute (need --read attribute)'),
parser.add_option('-b', '--boolean', dest='addbool', help='Add booleon (need --add attribute'),
parser.add_option('-o', '--object', action="store_true", default=False, dest='addobject', help='Add object (need --add attribute'),
(options, args) = parser.parse_args()

filename = ''
if options.filename != None:
	filename = options.filename
filterseparator = '_'

if options.addsol != None:
	if filename != '':
		if options.filtername != None:
			if options.createsol == True:
				content={'settings': ['object', {}]}
			else:
				readsolfile = solfile(filename)
				content = readsolfile.get_sol()
			contenttmp = content
			filtervals = options.filtername.split(filterseparator)
			for filterval in filtervals:
				contenttmp = contenttmp[filterval][1]

			if options.addbool != None:
				contenttmp[options.addsol] = ['boolean', options.addbool]
			elif options.addobject == True:
				contenttmp[options.addsol] = ['object', {}]
			else:
				print "no attr"
				sys.exit(1)
		else:
			print "no filter"
			sys.exit(1)
	else:
		print "no filename"
		sys.exit(1)
	readsolfile2 = solfile(filename)
	readsolfile2.set_sol(content)
	sys.exit()

if options.supprsol != None:
	if filename != '':
		if options.filtername != None:
			readsolfile = solfile(filename)
			content = readsolfile.get_sol()
			contenttmp = content
			filtervals = options.filtername.split(filterseparator)
			for filterval in filtervals:
				contenttmp = contenttmp[filterval][1]
			del contenttmp[options.supprsol]
		else:
			print "no filter"
			sys.exit(1)
	else:
		print "no filename"
		sys.exit(1)
	readsolfile2 = solfile(filename)
	readsolfile2.set_sol(content)
	sys.exit()

if filename != '':
	#filename = '.macromedia/Flash_Player/macromedia.com/support/flashplayer/sys/settings.sol'
	readsolfile = solfile(filename)
	if options.filtername != None:
		readsolfile.filter(options.filtername, filterseparator)
	content = readsolfile.get_sol()
	debugreadfile = readsolfile.print_debug()
	if options.debugit == True:
		print debugreadfile
	if options.testit == False:
		print content
	else:
		debugreadfile.sort()

		filename2 = 'test.sol'
		writesolfile = solfile(filename2)
		writesolfile.set_sol(content, 'yes')
		debugwritefile = writesolfile.print_debug()
		if options.debugit == True:
			print debugwritefile
		debugwritefile.sort()
		if debugreadfile == debugwritefile:
			print 'ok'
		else:
			print 'error with %s ! please send me this file !' % filename
			sys.exit(1)
